import java.util.ArrayList;
import java.util.LinkedList;

public class Terrain {
	private Point[][] terrain;
	
	public Terrain(ArrayList<Ridge> ridges) {
		this.terrain = this.generateTerrain(ridges);
	}
	
	public Terrain(double[][] heights) {
		this.terrain = this.generateTerrain(heights);
	}
	
	private Point[][] generateTerrain(ArrayList<Ridge> ridges) {
		Point[][] terrain = new Point[RescueOperation.TERRAIN_HEIGHT][RescueOperation.TERRAIN_WIDTH];
		
		for (int y = 0; y < terrain.length; y++) {
			for (int x = 0; x < terrain[0].length; x++) {
				double height = 0.0;
				for (Ridge r : ridges) {
					double d = DistanceCalculator.distanceFromPointToSegment(x, y, r.getP1X(), r.getP1Y(), r.getP2X(), r.getP2Y());
					
					double f = r.getHeight() * Math.pow(Math.E, -Math.pow(d, 2) / (2 * Math.pow(r.getStdev(), 2)) );
					
					height = Math.max(height, f);
				}
				terrain[y][x] = new Point(this, x, y, height);
			}
		}
		
		return terrain;
	}
	
	public Point[][] generateTerrain(double[][] heights) {
		Point[][] terrain = new Point[RescueOperation.TERRAIN_HEIGHT][RescueOperation.TERRAIN_WIDTH];
		
		for (int y = 0; y < terrain.length; y++) {
			for (int x = 0; x < terrain[0].length; x++) {
				terrain[y][x] = new Point(this, x, y, heights[y][x]);
			}
		}

		return terrain;
	}
	
	/*
	 * Returns list of points that are near (distance <= d) the segment [p, q].
	 */
	public ArrayList<Point> getPointsNearSegment(Point p, Point q, double d) {
		ArrayList<Point> points = new ArrayList<Point>();
		
		for (int y = 0; y < this.terrain.length; y++) {
			for (int x = 0; x < this.terrain[0].length; x++) {
				if (DistanceCalculator.distanceFromPointToSegment(x, y, p.getX(), p.getY(), q.getX(), q.getY()) 
						<= d) {
					points.add(this.terrain[y][x]);
				}
			}
		}
		
		return points;
	}
	
	/* Bresenham-based supercover line algorithm 
	 * for getting the grid squares between two points
	 * 
	 * http://lifc.univ-fcomte.fr/~dedu/projects/bresenham/index.html
	 */
	public ArrayList<Point> useVisionLine(Point p, Point q) { 
		ArrayList<Point> points = new ArrayList<Point>();
		int i;               // loop counter 
		int x1 = p.getX();
		int y1 = p.getY();
		int x2 = q.getX();
		int y2 = q.getY();
		int ystep, xstep;    // the step on y and x axis 
		int error;           // the error accumulated during the increment 
		int errorprev;       // *vision the previous value of the error variable 
		int y = y1, x = x1;  // the line points 
		int ddy, ddx;        // compulsory variables: the double values of dy and dx 
		int dx = x2 - x1; 
		int dy = y2 - y1; 
		points.add(p);  // first point 
		// NB the last point can't be here, because of its previous point (which has to be verified) 
		if (dy < 0) { 
			ystep = -1; 
			dy = -dy; 
		} else 
			ystep = 1; 
		if (dx < 0) { 
			xstep = -1; 
			dx = -dx; 
		} else 
			xstep = 1; 
		ddy = 2 * dy;  // work with double values for full precision 
		ddx = 2 * dx; 
		if (ddx >= ddy) {  // first octant (0 <= slope <= 1) 
			// compulsory initialization (even for errorprev, needed when dx==dy) 
			errorprev = error = dx;  // start in the middle of the square 
			for (i=0 ; i < dx ; i++) {  // do not use the first point (already done) 
				x += xstep; 
				error += ddy; 
				if (error > ddx) {  // increment y if AFTER the middle ( > ) 
					y += ystep; 
					error -= ddx; 
					// three cases (octant == right->right-top for directions below): 
					if (error + errorprev < ddx)  // bottom square also 
						points.add(this.terrain[y-ystep][x]); 
					else if (error + errorprev > ddx)  // left square also 
						points.add(this.terrain[y][x-xstep]);
					else {  // corner: bottom and left squares also 
						points.add(this.terrain[y-ystep][x]); 
						points.add(this.terrain[y][x-xstep]);
					} 
				} 
				points.add(this.terrain[y][x]); 
				errorprev = error; 
			} 
		} else {  // the same as above 
	    errorprev = error = dy; 
	    for (i=0 ; i < dy ; i++) { 
	    	 y += ystep; 
	    	 error += ddx; 
	    	 if (error > ddy) { 
	    		 x += xstep; 
	    		 error -= ddy; 
	    		 if (error + errorprev < ddy) 
	    			 points.add(this.terrain[y][x-xstep]);
	    		 else if (error + errorprev > ddy) 
	    			 points.add(this.terrain[y-ystep][x]); 
	    		 else { 
	    			 points.add(this.terrain[y][x-xstep]);
	    			 points.add(this.terrain[y-ystep][x]); 
	    		 } 
	    	 }
	    	 points.add(this.terrain[y][x]);
	    	 errorprev = error; 
	    	}
		}
		// assert ((y == y2) && (x == x2));  // the last point (y2,x2) has to be the same with the last point of the algorithm 
		return points;
	}
	
	/*
	 * Definition of feasibility:
	 * 
	 * Let h be some height. Let S be the set of grid points "near" the 
	 * segment pq (here "near" means eg within distance 1 or smth. like 
	 * that). Two points p and q are feasible if every point in S is seen 
	 * by either a helicopter at p or a helicopter at q flying at height h.
	 */
	public boolean isFeasible(Point p, Point q) {
		ArrayList<Point> points = this.useVisionLine(p, q);
		Helicopter h1 = new Helicopter(p, RescueOperation.HELICOPTER_HEIGHT);
		Helicopter h2 = new Helicopter(q, RescueOperation.HELICOPTER_HEIGHT);
		
		for (Point point : points) {
			if (!h1.isVisible(point) && !h2.isVisible(point)) {
				return false;
			}
		}
		
		return true;
	}
	
	/*
	 * Checks the feasibility of every pair of points in two tracks
	 * and returns the results in a 2D table of boolean values.
	 */
	public boolean[][] getFeasibilityTable(Track t1, Track t2) {
		boolean[][] feasibilities = new boolean[t2.getPoints().size()][t1.getPoints().size()];
		for (int i = 0; i < t1.getPoints().size(); i++) {
			for (int j = 0; j < t2.getPoints().size(); j++) {
				feasibilities[j][i] = this.isFeasible(t1.getPoints().get(i), t2.getPoints().get(j));
			}
		}
		return feasibilities;
	}
	
	/*
	 * 
	 */
	public String[] getFlyingDirections(ArrayList<Track> tracks) {
		String[] directions = new String[tracks.size() - 1];
		
		for (int t = 0; t < tracks.size() - 1; t++) {
			Track t1 = tracks.get(t);
			Track t2 = tracks.get(t + 1);
			directions[t] = this.getFlyingDirections(t1, t2);
			if (directions[t] == null) continue;
			directions[t] = directions[t].replaceAll("R", "f" + t + ",");
			directions[t] = directions[t].replaceAll("L", "b" + t + ",");
			directions[t] = directions[t].replaceAll("D", "f" + (t + 1) + ",");
			directions[t] = directions[t].replaceAll("U", "b" + (t + 1) + ",");
			directions[t] = directions[t].substring(0, directions[t].length() - 1);
		}
		
		return directions;
	}
	
	/*
	 * Finds the shortest path from (0, 0) to (t1.size-1, t2.size-1) in the
	 * feasibility table, using only feasible points. Returns a string of 
	 * chars containing the flying directions in the feasibility table:
	 * 
	 * R: x <- x+1      L: x <- x-1
	 * D: y <- y+1      U: y <- y-1
	 * 
	 */
	public String getFlyingDirections(Track t1, Track t2) {
		boolean[][] feasibilities = this.getFeasibilityTable(t1, t2);

		//this.print(feasibilities); System.out.println();
		
		int h = feasibilities.length;
		int w = feasibilities[0].length;
		char[][] tree = new char[h][w];
		java.awt.Point start = new java.awt.Point(0, 0);
		java.awt.Point end = new java.awt.Point(w - 1, h - 1);
    	LinkedList<java.awt.Point> queue = new LinkedList<java.awt.Point>();
    	LinkedList<java.awt.Point> marked = new LinkedList<java.awt.Point>();
    	LinkedList<java.awt.Point> path = new LinkedList<java.awt.Point>();
    	marked.add(start);
    	queue.add(start);
    	path.add(start);

		if (start.equals(end)) {
			if (this.isFeasible(t1.getPoints().get(0), t2.getPoints().get(0))) {
				return "";
			} else {
				return null;
			}
		}
		
		while (!queue.isEmpty()) {
			java.awt.Point t = queue.pop();
			
			for (int i = 1; i <= 4; i++) {
				int dx = 0;
				int dy = 0;
				char d;
				
				if (i == 1) {dx = -1; d = 'L';}
				else if (i == 2) {dx = 1; d = 'R';}
				else if (i == 3) {dy = -1; d = 'U';}
				else if (i == 4) {dy = 1; d = 'D';}
				else {d = '-';}
				
				java.awt.Point u = new java.awt.Point(t);
				u.translate(dx, dy);
				
				if (u.getX() < 0 || 
					u.getX() >= w || 
					u.getY() < 0 || 
					u.getY() >= h || 
					!feasibilities[(int) u.getY()][(int) u.getX()]) {
					continue;
				}
				
				if (u.equals(end)) {
					tree[(int) u.getY()][(int) u.getX()] = d;
					return this.getString(tree, start, end);
				}
				
				if (!marked.contains(u)) {
					marked.add(u);
					queue.add(u);
					tree[(int) u.getY()][(int) u.getX()] = d;
				}
			}
		}
		
		return null;
	}
	
	/*
	 * Helper method for getFlyingDirections.
	 */
	private String getString(char[][] table, java.awt.Point start, java.awt.Point end) {
		String string = "";
		java.awt.Point p = end;
		while (!p.equals(start)) {
			char d = table[(int) p.getY()][(int) p.getX()];
			int dx = 0;
			int dy = 0;
			string = d + string;
			if (d == 'U') dy = 1;
			else if (d == 'D') dy = -1;
			else if (d == 'L') dx = 1;
			else if (d == 'R') dx = -1;
			p.translate(dx, dy);
		}
		return string;
	}
	
	public Point getPoint(int x, int y) {
		if (x < 0 || y < 0 || x >= this.terrain[0].length || y >= this.terrain.length) {
			return null;
		} else {
			return this.terrain[y][x];
		}
	}
	
    private void print(boolean[][] table) {
    	for (int y = table.length - 1; y >= 0; y--) {
    		for (int x = 0; x < table[0].length; x++) {
    			char c;
    			if (table[y][x]) c = '1'; else c = '0';
     			System.out.print(c);
    		}
    		System.out.println();
    	}
    }
    
    public String toString() {
    	String s = "";
    	
    	for (int y = 0; y < this.terrain.length - 1; y++) {
    		for (int x = this.terrain[0].length - 1; x >= 0; x--) {
    			s += this.terrain[y][x].getZ();
    			if (x != 0) {
    				s += " ";
    			}
    		}
    		if (y != this.terrain.length - 1) {
    			s += "\n";
    		}
    	}
    	
    	return s;
    }

}
